/* Generic interface for Pattern */

#include <ctype.h>
#include <stdlib.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "Pattern.h"
#include "Pattern_p.h"




/* ---- NopPattern --------------------------------------------------------- */

static int
_NopPattern_initialize(Pattern* self, const char* arg)
{
	return 0;
}


static int
_NopPattern_match_wp(const Pattern* self, WPARAM wp)
{
	return !0;
}


static int
_NopPattern_match_lp(const Pattern* self, LPARAM lp)
{
	return !0;
}


static PatternType _NopPattern_Type = {
	"NopPattern",
	sizeof(Pattern),
	_NopPattern_initialize,
	Pattern_Finalize_NothingToDo,
	_NopPattern_match_wp,
	_NopPattern_match_lp,
	Pattern_CalcInstanceSize_Simple
};
static const PatternType* NopPattern_Type = &_NopPattern_Type;








/* ---- NumberPattern ------------------------------------------------------ */

typedef struct {
	Pattern_HEAD

	long value;
} NumberPattern;




static int
_NumberPattern_initialize(Pattern* _self, const char* arg)
{
	NumberPattern* self;

	self = (NumberPattern*)_self;

	self->value = strtol(arg, NULL, 0);

	return 0;
}


static int
_NumberPattern_match_wp(const Pattern* _self, WPARAM wp)
{
	NumberPattern* self;

	self = (NumberPattern*)_self;

	return wp == (WPARAM)(self->value);
}


static int
_NumberPattern_match_lp(const Pattern* _self, LPARAM lp)
{
	NumberPattern* self;

	self = (NumberPattern*)_self;

	return lp == (LPARAM)(self->value);
}


static PatternType _NumberPattern_Type = {
	"NumberPattern",
	sizeof(NumberPattern),
	_NumberPattern_initialize,
	Pattern_Finalize_NothingToDo,
	_NumberPattern_match_wp,
	_NumberPattern_match_lp,
	Pattern_CalcInstanceSize_Simple
};
static const PatternType* NumberPattern_Type = &_NumberPattern_Type;








/* ---- StringPattern ------------------------------------------------------ */

typedef struct {
	Pattern_HEAD

	char string[1];
} StringPattern;




static int
_StringPattern_initialize(Pattern* _self, const char* arg)
{
	StringPattern* self;

	self = (StringPattern*)_self;

	strcpy(self->string, arg);

	return 0;
}


static int
_StringPattern_match_wp(const Pattern* _self, WPARAM wp)
{
	StringPattern* self;

	self = (StringPattern*)_self;

	return strcmp(self->string, (const char*)wp) == 0;
}


static int
_StringPattern_match_lp(const Pattern* _self, LPARAM lp)
{
	StringPattern* self;

	self = (StringPattern*)_self;

	return strcmp(self->string, (const char*)lp) == 0;
}


static int
_StringPattern_calc_instance_size(const PatternType* type, const char* arg)
{
	return type->basic_size + strlen(arg) * sizeof(char);
}


static PatternType _StringPattern_Type = {
	"StringPattern",
	sizeof(StringPattern),
	_StringPattern_initialize,
	Pattern_Finalize_NothingToDo,
	_StringPattern_match_wp,
	_StringPattern_match_lp,
	_StringPattern_calc_instance_size
};
static const PatternType* StringPattern_Type = &_StringPattern_Type;








/* ---- StringBeginningPattern --------------------------------------------- */

typedef struct {
	Pattern_HEAD

	int length;
	char string[1];
} StringBeginningPattern;




static int
_StringBeginningPattern_initialize(Pattern* _self, const char* arg)
{
	StringBeginningPattern* self;

	self = (StringBeginningPattern*)_self;

	self->length = strlen(arg);
	strcpy(self->string, arg);

	return 0;
}


static int
_StringBeginningPattern_match_wp(const Pattern* _self, WPARAM wp)
{
	StringBeginningPattern* self;

	self = (StringBeginningPattern*)_self;

	return strncmp(self->string, (const char*)wp, self->length) == 0;
}


static int
_StringBeginningPattern_match_lp(const Pattern* _self, LPARAM lp)
{
	StringBeginningPattern* self;

	self = (StringBeginningPattern*)_self;

	return strncmp(self->string, (const char*)lp, self->length) == 0;
}


static PatternType _StringBeginningPattern_Type = {
	"StringBeginningPattern",
	sizeof(StringBeginningPattern),
	_StringBeginningPattern_initialize,
	Pattern_Finalize_NothingToDo,
	_StringBeginningPattern_match_wp,
	_StringBeginningPattern_match_lp,
	_StringPattern_calc_instance_size
};
static const PatternType* StringBeginningPattern_Type
                              = &_StringBeginningPattern_Type;








/* ---- BinaryPattern ------------------------------------------------------ */

typedef struct {
	Pattern_HEAD

	int offset;
	int length;
	unsigned char bytes[1];
} BinaryPattern;




/**
 * @offset,bytes
 *
 * @12,00 0f 0f ae ...
 * @12,000f0fae...
 */

static int
_xdigit_to_int(int c)
{
	if (isdigit(c))
		return c - '0';
	else if (islower(c))
		return c - 'a' + 10;
	else
		return c - 'A' + 10;
}

static int
_BinaryPattern_initialize(Pattern* _self, const char* arg)
{
	BinaryPattern* self;
	const char* p;
	int i;
	int f;

	self = (BinaryPattern*)_self;

	p = strchr(arg, ',');
	if (p == NULL)
		return !0;

	self->offset = atoi(arg);
	i = 0;
	f = 0;
	p++;
	while (*p != '\0') {
		if (isxdigit(*p)) {
			if (!f) {
				self->bytes[i] = _xdigit_to_int(*p);
				f = !0;
			} else {
				self->bytes[i] <<= 4;
				self->bytes[i] |= _xdigit_to_int(*p);
				f = 0;
				i++;
			}
		}
		p++;
	}

	if (f)
		return !0;

	self->length = i;
	return 0;
}


static int
_BinaryPattern_match_wp(const Pattern* _self, WPARAM wp)
{
	BinaryPattern* self;

	self = (BinaryPattern*)_self;

	return 0 == memcmp( self->bytes,
	                    ((const unsigned char*)wp) + self->offset,
	                    self->length );
}


static int
_BinaryPattern_match_lp(const Pattern* _self, LPARAM lp)
{
	BinaryPattern* self;

	self = (BinaryPattern*)_self;

	return 0 == memcmp( self->bytes,
	                    ((const unsigned char*)lp) + self->offset,
	                    self->length );
}


static int
_BinaryPattern_calc_instance_size(const PatternType* type, const char* arg)
{
	const char* p;
	int n;

	p = strchr(arg, ',');
	n = 0;
	if (p != NULL) {
		p++;
		while (*p != '\0') {
			if (isxdigit(*p))
				n++;
			p++;
		}
	}

	return type->basic_size + (n / 2) * sizeof(unsigned char);
}


static PatternType _BinaryPattern_Type = {
	"BinaryPattern",
	sizeof(BinaryPattern),
	_BinaryPattern_initialize,
	Pattern_Finalize_NothingToDo,
	_BinaryPattern_match_wp,
	_BinaryPattern_match_lp,
	_BinaryPattern_calc_instance_size
};
static const PatternType* BinaryPattern_Type = &_BinaryPattern_Type;








/* ---- Generic interface for Pattern -------------------------------------- */


static Pattern*
_Pattern_New(const PatternType* type, const char* arg)
{
	Pattern* self;

	self = malloc((*(type->calc_instance_size))(type, arg));
	if (self == NULL) return NULL;

	self->type = type;
	if ((*(type->initialize))(self, arg) != 0) {
		free(self);
		return NULL;
	}

	return self;
}


void
Pattern_Del(Pattern* self)
{
	(*(self->type->finalize))(self);
	free(self);
}




#define IS_NOP_PATTERN(s) ((s)[0] == '-' && (s)[1] == '\0')
#define IS_NUMBER_PATTERN(s) ( isdigit((s)[0]) \
                               || ((s)[0] == '-' && isdigit((s)[1])) )
#define IS_STRING_PATTERN(s) ((s)[0] == '"')
#define IS_STRING_BEGINNING_PATTERN(s) ((s)[0] == '^')
#define IS_BINARY_PATTERN(s) ((s)[0] == '@')

Pattern*
Pattern_Compile(const char* pattern_string)
{
	const PatternType* type;

	if (IS_NOP_PATTERN(pattern_string))
		type = NopPattern_Type;
	else if (IS_NUMBER_PATTERN(pattern_string))
		type = NumberPattern_Type;
	else if (IS_STRING_PATTERN(pattern_string)) {
		type = StringPattern_Type;
		pattern_string += 1;
	} else if (IS_STRING_BEGINNING_PATTERN(pattern_string)) {
		type = StringBeginningPattern_Type;
		pattern_string += 1;
	} else if (IS_BINARY_PATTERN(pattern_string)) {
		type = BinaryPattern_Type;
		pattern_string += 1;
	} else
		return NULL;

	return _Pattern_New(type, pattern_string);
}




int
Pattern_MatchWparam(const Pattern* self, WPARAM wp)
{
	return (*(self->type->match_wp))(self, wp);
}


int
Pattern_MatchLparam(const Pattern* self, LPARAM lp)
{
	return (*(self->type->match_lp))(self, lp);
}




void
Pattern_Finalize_NothingToDo(Pattern* self)
{
	return;
}


int
Pattern_CalcInstanceSize_Simple(const PatternType* type, const char* arg)
{
	return type->basic_size;
}




/* __END__ */
